#include <stdio.h>                  // printf()
#include <dos.h>                    // geninterrupt()
#include "..\types\types.h"
#include "..\dpmi32\realmode.h"



// returns the physical address of a real-mode segment:offset pointer
//
// the reverse mapping is not so simple, because (1) we need
// to return an error when physical address is greater than
// 1 MB, (2) we need to return 2 separate values for segment
// and offset, and (3) there are many real-mode mappings for
// a given physical address
//
Bit32 real_to_physical( Bit32 segment_offset )
{
    return ((segment_offset >> 12) & 0xFFFF0) + (segment_offset & 0xFFFF);
}



// allocates a block of memory with a physical address below
// 1 MB line, so real-mode programs can use it
//
// returns both the real mode segment and the protected mode
// selector; note that accessing the memory from 0:32 flat
// protected mode will require an additional call to map_physical()
// unless you want to use far_read32() and far_write32() functions
// each time, to access the memory indirectly
//
//
//
// does not care about current contents of selector
//
// if an error is returned, none of the memory has been allocated
//
// length_available is only filled in when an error is returned
//
// since we are a 32-bit client, only one descriptor is ever
// allocated, so there is no need to mess about incrementing
// the selector by the amount we get from DPMI function 0x0003
//
// you must not modify or free the descriptors returned by this
// function, except with free_dos_mem() and realloc_dos_mem()
//
// uses DOS INT 0x21 function 0x48
//
// possible errors:
//      0x0007      memory control blocks damaged
//      0x0008      insufficient memory
//      0x8011      descriptor unavailable
//
//      TRAP 0D     you passed 0 length
//
Bit16 alloc_dos_mem( Bit32 length_wanted, Bit32 *length_available, Bit16 *selector, Bit16 *segment )
{
    union REGS regs;
    Bit32 wide;

    // round up to the next paragraph (16 bytes)
    //
    // max out at 64K paragraphs, rather than passing
    // only the bottom 16 bits, to force an error
    //
    wide = (length_wanted + 15) >> 4;
    if( wide > 65535 )
        wide = 65535;
    regs.w.bx = wide;

    regs.w.ax = 0x0100;
    regs.w.flags = 0;
    int386( 0x31, &regs, &regs );

    if( regs.w.cflag )
    {
        *length_available = regs.w.bx << 4;
        return regs.w.ax;
    }

    *segment = regs.w.ax;
    *selector = regs.w.dx;

    return 0;
}



// changes the size of a memory block previously allocated by
// alloc_dos_mem() or realloc_dos_mem()
//
//
//
// expects selector to already be valid, and alters caller's
// variable if successful
//
// increasing the size of a DOS memory block may fail even if
// the original allocation would have succeeded with this size,
// because subsequent DOS allocations (either from protected
// mode or directly from real mode) have fragmented DOS memory
//
// since we are a 32-bit client, only one LDT descriptor is ever
// needed, so we don't have to worry about increasing size failing
// because the next entry in the LDT is unavailable
//
// since we are a 32-bit client, only one LDT descriptor is ever
// needed, we don't have to worry about decreasing size causing
// some descriptors being deallocated
//
// last descriptor for the block (only one since we are 32-bit
// client) will have its limit adjusted
//
// under DPMI 1.0 hosts, any (protected mode) segment registers
// which contain a selector being freed are set to zero
//
// possible errors:
//      0x0007      memory control blocks damaged
//      0x0008      insufficient memory
//      0x0009      incorrect memory segment specified
//      0x8011      descriptor unavailable
//      0x8022      invalid descriptor
//
Bit16 realloc_dos_mem( Bit32 length_wanted, Bit32 *length_available, Bit16 *selector, Bit16 *segment )
{
    union REGS regs;

    // round up to the next paragraph (16 bytes)
    //
    regs.w.bx = (length_wanted + 15) >> 4;

    regs.w.dx = *selector;
    regs.w.ax = 0x0102;
    regs.w.flags = 0;
    int386( 0x31, &regs, &regs );

    if( regs.w.cflag )
    {
        *length_available = regs.w.bx << 4;
        return regs.w.ax;
    }

    *segment = regs.w.ax;
    *selector = regs.w.dx;

    return 0;
}



// frees a memory block previously allocated by alloc_dos_mem()
// or realloc_dos_mem()
//
//
//
// since we are a 32-bit client, only one descriptor was ever
// allocated, so there is no issue with passing the correct one
// or remembering that more than one will be freed by this call
//
// under DPMI 1.0 hosts, any (protected mode) segment registers
// which contain a selector being freed are set to zero
//
// possible errors:
//      0x0007      memory control blocks damaged
//      0x0009      incorrect memory segment specified
//      0x8022      invalid descriptor
//
Bit16 free_dos_mem( Bit16 selector )
{
    // must assign to EBX or one of its pieces in order to get
    // the compiler to preserve its previous value, because the
    // calling convention puts the onus on us, and the interrupt
    // trashes EBX (found this out the hard way!)
    //
    _EBX = 0x12345678;

    _DX = selector;
    _AX = 0x0101;
    geninterrupt( 0x31 );

    if( _FLAGS & 1 )
        return _AX;

    return 0;
}



// returns a unique segment:offset that real mode code can do
// a far call to transfer to a protected mode procedure
//
// when called, interrupts will be disabled, you will be using
// a locked protected mode stack, and registers have these values:
//
//      DS:ESI      protected mode alias for real mode SS:SP
//      ES:EDI      protected mode pointer to RealModeRegs structure
//
// when done, your code should execute an IRET with ES:EDI pointing
// at a RealModeRegs data structure (although not necessarily the
// same one you got on entry)
//
// beware of reentrancy -- the protected mode alias for the real
// mode stack, and the real mode register structure are statically
// allocated, so being called back when you are still in the middle
// of a callback will overwrite your state unless you save them
// before doing anything that could cause your code to be called
// again, such as enabling interrupts or calling some real mode
// code explicitly
//
// you will need some assembly language thunking code to push the
// parameter registers onto the stack for use by your C code and
// reload DS, but it doesn't need to be as extensive as when you
// are handling a hardware interrupt, since you don't have to
// preserve any registers
//
// if you really want to be cheap, you can ignore the RealModeRegs
// protected mode pointer that DPMI loaded into ES:ESI, since that
// will be the same as you are passing this function now
//
// your C code will also need to access the passed stack selector via
// far_read32() and far_write32() unless you want to call map_physical()
// and unmap_physical(), which may be too slow to do every time you get
// called from real mode
//
// to access any data which is pointed to by the real mode registers
// or real mode stack, you will definitely need to call real_to_physical()
// and map_physical()
//
//
//
// note that even though you provide the pointer to the RealModeRegs
// structure, it is not filled in until you get called from real mode
//
// DPMI host is required to provide at least 16 callback addresses
// per client; host may also allocate one descriptor per callback
// on your behalf to hold real mode stack alias
//
// possible errors:
//      0x8015      callback unavailable
//
static Bit16 alloc_real_callback_helper(
    vfp protmode_function,
    struct RealModeRegs *real_regs,
    Bit32 *real_segment_offset
)
{
#if 0
    union REGS regs;
    struct SREGS sregs;

    // we must use int386x() rather than assigning directly
    // to registers ourselves, because we cannot touch segment
    // registers safely
    //
    // unfortunately, int386x() loads all segment registers,
    // including DS, before dereferencing the regs pointer,
    // so it ends up reading (before the interrupt) and writing
    // (after the interrupt) the regs structure using the offset
    // you pushed onto the stack, but the selector in the SREGS
    //
    // so if we pass a code selector in DS register, we get a
    // trap D when int386x() tries to save the EAX register it
    // got back from the interrupt
    //
    // so I'm gonna try passing _DS as the code segment for the
    // flat mode routine to execute, but I'm afraid I am only
    // deferring another TRAP D when it tries to execute the
    // callback code...
    //

    sregs.fs = _FS;
    sregs.gs = _GS;

//  sregs.ds = _CS;
    sregs.ds = _DS;     // probably wrong, but it's my only damn choice!

    regs.x.esi = (Bit32) protmode_function;
    sregs.es = _DS;
    regs.x.edi = (Bit32) real_regs;
    regs.w.ax = 0x0303;
    regs.w.flags = 0;
    int386x( 0x31, &regs, &regs, &sregs );

    if( regs.w.cflag )
        return regs.w.ax;

    *real_segment_offset = (regs.w.cx << 16) | regs.w.dx;
#else
    // we had better touch the EBX register so the
    // C compiler will preserve it for our caller,
    // in case the DPMI call alters it
    //
    _EBX = 0x12345678;

    _ESI = (Bit32) protmode_function;
    _EDI = (Bit32) real_regs;

    // i'm gonna play fast and loose with segment registers
    // and see what happens!
    //
    // assumes ES starts out the same as DS register
    //
    _DS = _CS;

    _AX = 0x0303;
    geninterrupt( 0x31 );


    // assumes ES starts out the same as DS register
    //
    // and since the compiler uses AX register as a
    // temporary, assumes we don't care about value
    // in AX after this system call
    //
    _DS = _ES;

    if( _FLAGS & 1 )
        return _AX;

    *real_segment_offset = (_CX << 16) | _DX;
#endif

    return 0;
}




#ifdef __cplusplus
extern "C"
#endif
void real_callback_thunk( void );



// deals with the thunking issues of having a C function called
// back by real mode code
//
// expects space for an extra pointer just before the RealModeRegs
// data structure, so we can store the pointer to the user's C
// function there
//
// the protected mode procedure passed called by DPMI will always
// be the thunking code defined in REAL2.ASM
//
Bit16 alloc_real_callback(
    cfp protmode_c_function,
    Bit8 *protmode_c_stack,
    struct RealModeRegsC *real_regs_c,
    Bit32 *real_segment_offset
)
{
    real_regs_c->stack = protmode_c_stack;
    real_regs_c->function = protmode_c_function;
    return alloc_real_callback_helper( real_callback_thunk, &real_regs_c->regs, real_segment_offset );
}



// frees what you got from alloc_real_callback()
//
//
//
// possible errors:
//      0x8024      invalid callback address
//
Bit16 free_real_callback( Bit32 real_segment_offset )
{
    _CX = real_segment_offset >> 16;
    _DX = real_segment_offset & 0xFFFF;
    _AX = 0x0304;
    geninterrupt( 0x31 );

    if( _FLAGS & 1 )
        return _AX;

    return 0;
}



// deals with the thunking issues of calling a real mode software
// interrupt handler from protected mode, particularly when the
// real mode code wants far pointers, which means loading DS and
// registers with values which are invalid in protected mode
//
// copying words from protected mode stack to real stack is not
// implemented; maybe this needs to be assembly code to do that???
// but if we did it, it would be the caller's responsibility to
// remove parameters from the protected mode stack
//
// alters the caller's RealModeRegs structure after interrupt is done
//
// caller must initialize the reserved field in RealModeRegs to 0
//
// if caller sets SS:ESP to 0:0 the DPMI host will provide a real
// mode stack
//
// IP and CS in RealModeRegs will be ignored, since they are derived
// from interrupt_number
//
// the flags in RealModeRegs will be pushed onto the real mode stack's
// IRET frame, but the interrupt handler will be called with interrupt
// and trace flags clear
//
// obviously the segment registers need to be valid in real mode
//
// the target real mode handler must return on the same stack it was
// called on, using the IRET instruction
//
// possible errors:
//      0x8012      linear memory unavailable (stack)
//      0x8013      physical memory unavailable (stack)
//      0x8014      backing store unavailable (stack)
//      0x8021      invalid value (asked to copy too much stack)
//
Bit16 simulate_real_interrupt( Bit8 interrupt_number, struct RealModeRegs *real_regs )
{
    // must assign to ESI or one of its pieces in order to get
    // the compiler to preserve its previous value, because the
    // calling convention puts the onus on us, and the interrupt
    // trashes ESI (found this out the hard way!)
    //
//    _ESI = 0x12345678;

    // no need to change ES register since in
    // flat protected mode it already points at
    // our C data segment
    //

    _EDI = (Bit32) real_regs;
    _CX = 0;                // no stack words copied
    _BH = 0;                // all bits reserved
    _BL = interrupt_number;
    _AX = 0x0300;
    geninterrupt( 0x31 );

    if( _FLAGS & 1 )
        return _AX;

    return 0;
}
