/* conditional compile switches all go here
 *
 * manifest constants are still sprinkled
 * all through the code (sorry)
 */
#define STUB_2ND_LEVEL_STATS    FALSE
#define ASSUME_OBJECT_IS_PQ     TRUE
#define ASSUME_ETHERTYPE_IS_IP  FALSE
#define TIME_ALREADY_SWAPPED    TRUE
#define BYTESWAP_FIELDS_EARLY   FALSE
#define ALWAYS_DUMP_HASH_INFO   FALSE
#define FOLD_BEFORE_MODULUS     TRUE
#define FOLD_RANDOMIZES         TRUE
#define FOLD_KEEPS_16_BITS      TRUE
#define USE_OLD_HASH            FALSE
#define BUFFER_USED_PARANOIA    FALSE
#define BUFFER_PTR_PARANOIA     FALSE
#define FLOAT_EXCEPT_PARANOIA   FALSE
#define DEBUG_CHAIN_LENGTH      FALSE
#define DEBUG_HASHINDEX         FALSE
#define DEBUG_COPY_PACKET       FALSE           // stupid debugger!
#define DEBUG_BUFFERING         TRUE
#define DEBUG_FILLIN            FALSE
#define DEBUG_HOUSECLEAN        FALSE
#define FAIL_ON_PACKET_COUNT    FALSE
#define TALK_WHEN_ALLOC_FAILS   FALSE           // for debugging when overflow occurs only!  too slow
#define DELTA_TIME_PARANOIA     FALSE           // worry about negative delta time since last flows_expire()
#define DELTA_TIME_FIXUP        FALSE           // truncate large time since last flows_expire() to 1 second
#define DELTA_TIME_STATS        TRUE            // track min/max/mean for delta time since last flows_expire()
#define DEBUG_QUERY_TOTALS      FALSE
#define DEBUG_MAX_BPS           FALSE
#define MAX_BPS_PARANOIA        TRUE
#define QUERY_CALLS_EXPIRE      TRUE
#define DEBUG_ACTIVE_FEED       FALSE
#define DEBUG_ACTIVE_FEED_DATA  FALSE
#define EXPIRE_OVERHEAD_STATS   TRUE
#define REPORT_OVERHEAD_STATS   FALSE



#define LARGEST_IP_SIZE         1500
#define UDP_PAYLOAD_SIZE        (LARGEST_IP_SIZE - 2*sizeof(struct ip) - sizeof(struct udphdr))
#define FLOWS_PER_UDP           (UDP_PAYLOAD_SIZE / sizeof(struct flowstats))
#define EXPECTED_PACKETS        155000          // per second (2 cells/packet, 1 occupied link)
#define ACTIVE_SPEW_MILLISECS   1000     // how often should we spew active flows? (milliseconds)

#define CONTROL_TCP_PORT           23

#define QUERY_TCP_PORT             22   // use secure shell (ssh) port to get past firewall
//#define QUERY_TCP_PORT           23   // use telnet port for now

#define QUERY_SOCKET_COUNT          2   // how many sockets to keep around

#define QUERY_BUFFER_SIZE       (3*1024*1024/2)     // 1.5 M bytes each






/*
OVERVIEW

this module is designed to be integrated into a
data capture system and called in realtime to convert
blocks of raw packet traces into "flows"

when we determine that a flow has finished (we haven't
seen any packets which hash the same in a user-defined
number of seconds) we pass it off to the statistics
module for further processing

occasionally, we receive a message on an out-of-band
channel (Ethernet when we are monitoring ATM) and we
tell the statistics module to finalize any info it has
for the flows we have already given it, and we send
those results back to whoever queried us, and then we
reset the statistics module




CREDITS

Written by Kevin Fall (kfall@sdsc.edu) 4/94 to 5/94
based on an earlier program by Hans-Werner Braun

change 4 mar 94 to verify whether protocol TCP (6) or UDP (17) and just
proceeds is same is the case. If not, it verifies whether the protocol
is ICMP (1), in which case only the two eight bit Type/Code fields are
retained. With all other protocols no information above the IP layer
is retained.

modified 12/5/95 by Joel Apisdorf:

to do hash chain length and concurrent flow count statistics

also changed hash algorithm to use a smaller, prime, number of buckets
and to use XOR and modulus to include all bits of all fields, rather
than using low 4 bits of each field and not combining them

also added -k command line parameter
so fillin messages can be turned off

also added -p command line parameter
so new flows messages can be turned off

modified 2/15/96 by Joel Apisdorf to use blocks of 128 packets
in the format captured by OC3MON.  Also changed the code
which ignores MAC-layer packets and non-first IP fragments to
understand the MAC-layer representation, rather than expecting
the fake flags seen in FR (FDDI receive) -generated traces.

modified 3/8/96 by Joel Apisdorf to coexist in same executable
and be called by data capture module, rather than be driven by
trace file on disk

modified 3/12/96 by Joel Apisdorf to support multiple queries
within the same run, so flows and stats can run continuously
without losing data, and so interval is controlled remotely,
like the NNSTAT agents

modified 10/30/96 by Kevin Thompson for a new option to spit
out active flow updates based on activity. Uses a 3rd linked list

*/



#include <sys/types.h>

#if __unix__
#   include <sys/time.h>
#   include <sys/resource.h>
#   include <netinet/in.h>
#else
#   include "..\wattcp\tcp.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <float.h>              /* DBL_MIN DBL_MAX */

#if __unix__
#   include "types.h"
#else
#   include "..\types\types.h"
#   include "..\intel\timestmp.h"
#   include "..\intel\byteswap.h"
#   include "..\intel\fpu.h"    /* get_fpu_status() clear_fpu_status() */
#endif

#include "queue.h"
#include "trace.h"              /* must be included before util.h */
#include "util.h"
#include "dostats.h"
#include "buffer.h"             /* class BufferBlock */
#include "fs2flows.h"           /* make sure we comply with our own prototypes */
#include "portstat.h"           /* porthash_alloc_failures      (that module should print this itself, but I don't want to break the Perl scripts just now) */
#include "hoststat.h"           /* hosthash_alloc_failures      (that module should print this itself, but I don't want to break the Perl scripts just now) */
#include "netstats.h"           /* nethash_alloc_failures       (that module should print this itself, but I don't want to break the Perl scripts just now) */



// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS



/*
 * user-selected object default
 */

#define DEFAULT_USER_PREF   "pq"


/*
 * how long to wait to time a flow out
 */

#define DEFAULT_FLOW_ACCESS_TIMEOUT      64.0   /* 64 seconds */
#define DEFAULT_FLOW_CREATE_TIMEOUT    3600.0   /* 1 hour */
static CardClock flow_access_timeout = DEFAULT_FLOW_ACCESS_TIMEOUT * CARD_CLOCK_RATE;
static CardClock flow_create_timeout = DEFAULT_FLOW_CREATE_TIMEOUT * CARD_CLOCK_RATE;


/*
 * housekeeping interval
 */

#define DEFAULT_HOUSEKEEP   1.0  /* seconds */
#define TVMULT          1.5 /* multiplier for "fillins" */
static CardClock housekeep_timeout = DEFAULT_HOUSEKEEP * CARD_CLOCK_RATE;

/* global status info
 */

#if DEBUG_HASHINDEX
    u_int hashindex_trace = -1;
#endif

u_int ip_length_usage[ IP_LEN_COUNT ];
u_int ip_precedence_usage[ IP_PREC_COUNT ];
Bit64 sum_bytes_per_packet[ MAX_ATM_NICS ];
Bit16 max_bytes_per_packet[ MAX_ATM_NICS ];
u_int sum_packets[ MAX_ATM_NICS ];
u_int max_packets[ MAX_ATM_NICS ];
u_int cleaned_packets[ MAX_ATM_NICS ];
Bit64 cleaned_bytes[ MAX_ATM_NICS ];
u_int cleaned_flows;
u_int total_packets_trace = -1;
u_int total_packets[ MAX_ATM_NICS ];
u_int non_ip_packets[ MAX_ATM_NICS ];
u_int ip_optioned_packets[ MAX_ATM_NICS ];
u_int non_first_ip_fragments[ MAX_ATM_NICS ];
u_int first_ip_fragments[ MAX_ATM_NICS ];
Bit64 total_bytes[ MAX_ATM_NICS ];
u_int sum_new_flows;
Bit64 sum_bytes[ MAX_ATM_NICS ];
Bit64 max_bytes[ MAX_ATM_NICS ];
Bit16 instant_max_bytes[ MAX_ATM_NICS ];
u_int sum_flows;
u_int max_flows;
u_int max_new_flows;
int report_already_set;
int istamp_already_set;
int last_already_set[ MAX_ATM_NICS ];
u_int query_count;                /* how many times have we been queried? */
u_int total_flows;                /* total ever processed */
u_int deleted_flows;              /* total ever deleted */
u_int ndeleted_before_clean;      /* tot deleted before cleanup */
static CardClock istamp;          /* time on 1st record */
static CardClock tstamp;          /* current time stamp */

#if EXPIRE_OVERHEAD_STATS
    u_int sum_deleted_flows;
    u_int max_deleted_flows;
#endif

#if DELTA_TIME_PARANOIA
    static CardClock prev_stamp;
    static CardClock beg_block_stamp;
    static Bit16 current_card, last_card, prev_card;
    static Bit16 beg_block_card, end_block_card;
#endif
#if DELTA_TIME_STATS
    static double min_houseclean_interval;
    static double max_houseclean_interval;
    static double sum_houseclean_interval;
#endif

/* stuff related to command line flags */
u_int pflag = FALSE, rflag = FALSE; /* cmd line flags */
u_int active_feed = FALSE; /* klt - 10/30/96 */
u_int fillin_wanted = TRUE;
int objtype;
char *objname = DEFAULT_USER_PREF;
CardClock end_block_stamp[ MAX_ATM_NICS ];
CardClock last_report_stamp[ MAX_ATM_NICS ];
CardClock next_report_stamp;
#if __unix__
    struct rusage before, after;
#else
    PentiumClock before, after;
#endif

/* TCP/IP socket stuff
 */
tcp_Socket control_socket;
tcp_Socket query_sockets[ QUERY_SOCKET_COUNT ];
Bit8 *query_output_buffers[ QUERY_SOCKET_COUNT ];
unsigned which_query_socket_to_use;

/* must measure time until socket is closed,
 * not time until flows_query() returns,
 * to know how slow network was, now that
 * TCP/IP stack buffers the entire report
 */
static unsigned which_query_socket_prev_used;
static PentiumClock query_start_time;
static double query_drainage_delta;

struct user_prefs {
    struct flowdata up_fd;  /* not completely used, but handy */
} prefs;

#define up_src      up_fd.fd_src
#define up_dst      up_fd.fd_dst
#define up_sport    up_fd.fd_sport
#define up_dport    up_fd.fd_dport

/*
 * general hashing routines
 *
 */


#if USE_OLD_HASH
#define NFLOWBUCKETS    (1024*1024)
#define flowhash(fd) ( (((u_int)(fd)->fd_src)&0xf << 16) | \
                  (((u_int)(fd)->fd_dst)&0xf << 12) | \
                  (((u_int)(fd)->fd_proto)&0xf << 8) | \
                  (((u_int)(fd)->fd_sport)&0xf << 4) | \
                  (((u_int)(fd)->fd_dport)&0xf) )
#else
/*
#define NFLOWBUCKETS    104729
#define NFLOWBUCKETS    32749
#define NFLOWBUCKETS    8191
#define NFLOWBUCKETS    4001
#define NFLOWBUCKETS    1021
#define NFLOWBUCKETS    251
*/
#define NFLOWBUCKETS    65536
#define BUCKET_POOL_SIZE (NFLOWBUCKETS * sizeof(struct hashrec *))
static struct hashrec **flowtab = NULL;

#if DEBUG_CHAIN_LENGTH
#   define CHAIN_POOL_SIZE (NFLOWBUCKETS * sizeof(int))
    static int *chainlen = NULL;
    static int longest_chain = 0;
#endif

#if !FOLD_BEFORE_MODULUS
#   define fold(x) (x)
#   define modulus(x) ((x) % NFLOWBUCKETS)
#else
#if FOLD_RANDOMIZES
#   if BIT64_AVAILABLE
#       define fold32(x) (((x)>>32) ^ (x))
#       define modulus(x) ((x) & 65535)
#   else
#       define fold32(x) (x)
#       define modulus(x) (x)
#   endif
#   define fold(x) ((fold32(x)*1664525 + 1013904223) >> 16)
#else
#if FOLD_KEEPS_16_BITS
#   define fold(x) (((x)>>16) ^ (x))
#   define modulus(x) ((x) & 65535)
#else
#   define fold(x) (((x)>>24) ^ ((x)>>16) ^ ((x)>>8) ^ (x))
#   define modulus(x) ((x) & 255)
#endif
#endif
#endif

#if BIT64_AVAILABLE
/* generate a 64-bit hash key
 */
#define flowhash(fd) modulus( fold( \
                  (*(Bit64 *)(fd)) ^ \
                  (*(Bit64 *)(8 + (Bit8 *)(fd)) >> 16) \
                  ) )
#else
/* generate a 32-bit hash key
 */
#define flowhash(fd) modulus( fold( \
                  ((Bit32)(fd)->fd_src.s_addr) ^ \
                  ((Bit32)(fd)->fd_dst.s_addr) ^ \
                  (*(Bit32 *)(12 + (Bit8 *)(fd))) ^ \
                  ((Bit32)(fd)->fd_proto) \
                  ) )
#endif      /* generate 32 bit key */
#endif      /* use new hash */


//#define HASH_RECORDS 70000L       // fills after 40 seconds at core.wtn 9/19/96 4:00 PM
//#define HASH_RECORDS 100000L      // fills after 1'2" at core.wtn 9/19/96 4:07 PM
//#define HASH_RECORDS 140000L      // not full after 14'30" at core.wtn 9/19/96 4:23P but overflowed 10/5/96 3:10P - 3:25P
//#define HASH_RECORDS 200000L      // too little 10/22/96 4:34P
//#define HASH_RECORDS 400000L      // not even 375K + 60K fit when 32 MB allocated to DPMI server
//#define HASH_RECORDS 350000L      // fills at core.wtn 1/7/97 1:05 PM
//#define HASH_RECORDS 590000L      // 419K needed at core.wtn 1/24/97 11:50 AM
//#define HASH_RECORDS 550000L      // larger structure, cannot increase mem avail to malloc
//#define HASH_RECORDS 545000L      // fills at core.wtn 2/11/97 all day
#define HASH_RECORDS 650000L        // how long will this last at WTN?
#define HASH_POOL_SIZE (HASH_RECORDS * sizeof(struct hashrec))
struct hashrec *hash_free = NULL;
struct hashrec *hash_pool = NULL;
u_int hash_alloc_failures;

/* klt - global stuff for active feed */
    u_char mttl = 255;    /* jza - 11/13/96 multicast time to live */
    struct in_addr taddr = {0}; /* jza - 11/13/96 address to tunnel multicast through */
    struct in_addr maddr; /* klt - 10/30/96 mcast address */
    udp_Socket msock;   /* klt - 11/10/96 the socket we'll get on a udp_open */
#define ACTIVE_FEED_PORT 6507   // klt - 11-10-96 port to send active feed stuff out to
/* */


int new_record( u_int indx, flowdata *fd, CardClock timestamp, Bit8 interface );
void delete_rec();
void usage( char *prog );
void summarize();
void summarize_rec( char *cp, struct user_prefs *up, struct hashrec *hr );
int flows_clear( void );


#if !__unix__
/* UNIX-compatible command-line options processing
 */
char *optarg = 0;
unsigned optcount = 0;

int getopt(
    int count,
    char **strings,
    char *pattern
)
{
    char *patpos;
    int result;

    for( ; optcount<count; optcount++ )
    {
        // does the current arg start with dash?
        //
        if( strings[optcount][0] == '-' )
        {
            optarg = strings[optcount]+1;
            for( patpos=pattern; patpos; patpos++ )

                // if this is an option the caller wants...
                //
                if( *optarg == *patpos )
                {
                    // remember the option character
                    // so we can return it
                    //
                    result = *optarg;

                    // move forward to the next argument
                    // so we know our place next time
                    //
                    optcount++;

                    // if caller wants a value attached to this arg...
                    //
                    patpos++;
                    if( *patpos == ':' )
                    {
                        // use up an argument
                        //
                        optarg = strings[ optcount ];
                        optcount++;
                    }

                    return result;
                }
        }
        // else ignore the argument completely
        //
    }

    optcount = 0;
    return -1;
}
#endif  /* not unix */




/*
 * do_record
 *  1> look up record... if there update access time
 *  2> if not there, insert it
 */

int do_record( struct tracerec *rp, CardClock timestamp, Bit8 interface )
{
    u_int hashindex;
    struct hashrec *hp;
    struct flowdata fd;

    /*
     * set fields appropriately, based on user prefs
     * must be done before hash index is computed
     */

    (*objtypes[objtype].update)(rp, &fd);

    hashindex = flowhash(&fd);

    if( hashindex >= NFLOWBUCKETS )
        printf( "internal error: flowhash %u > %d\n", hashindex, NFLOWBUCKETS );

    hp = flowtab[hashindex];

#if !ALWAYS_DUMP_HASH_INFO
    if( total_packets[ interface ] == total_packets_trace )
#endif
        fprintf( stderr, "interface %d packet %d hash index 0x%8.8lX from IP=%8.8X:%8.8X proto=%d TCP=%4.4X:%4.4X\n",
            interface, total_packets[interface], hashindex, fd.fd_src, fd.fd_dst, fd.fd_proto, fd.fd_sport, fd.fd_dport );

    if (pflag && rflag && flowcmp(&prefs.up_fd, &fd)) {
        CardClock tmp = timestamp - istamp;
        printf("FlowTrace:\t%.6f\t%d\t%x:%x\t%d:%d:%d\t%.6f\n",
            tmp.get_seconds(),
            fd.fd_len,
            fd.fd_src,
            fd.fd_dst,
            fd.fd_proto,
            fd.fd_sport,
            fd.fd_dport,
            timestamp.get_seconds()
        );
    }

    /* look for match or collision on hash chain
     */
    while( hp != NULL )
    {
        // if we found an existing flow...
        //
        int cmp = flowcmp(&hp->hr_fd, &fd);
        if( cmp )
        {
            // update the fields of the old record
            //
            hp->hr_packets++;
            hp->hr_bytes += Bit32( fd.fd_len );

            // deal with out-of-order packets...
            //
            // even if all the ATM cards' clocks were in perfect
            // synchronization, and the packets got copied to
            // host memory in the correct order, the blocks that they
            // were part of may have completed in a different order
            //
            // you cannot fix the packet ordering by simply swapping
            // the order of the blocks in the filled list, because
            // other packets in the blocks want the blocks to be in
            // a different order, as evidenced by the packets at the
            // end of the block that got us into this mess
            //
            // when we monitor a router's only link to the ATM switch,
            // and it needs to forward a packet to another ATM-connected
            // device, the packet appears on TX fiber just after RX fiber,
            // so they are likely to be in blocks that complete near each
            // other, causing the above problem
            //
            // of course, when you see a forwarded packet twice, it and
            // its bytes should only be counted once, which negates any
            // concern about which timestamp to touch, but I haven't
            // figured out a way to detect such a beast without being
            // dependent on a particular protocol's sequence numbering!
            //
            if( timestamp < hp->hr_start )
            {
                hp->hr_start = timestamp;
                hp->hr_start_if = interface;
            }
            else
            {
                hp->hr_access = timestamp;
                hp->hr_access_if = interface;
            }

            // move the flow to the end of the
            // time-sorted queue of access times
            //
            // note that since packets could be out-of-order,
            // endque() ought to be stepping back through linked list
            // to make sure that this gets inserted where its access
            // time is less than next in list, but that isn't
            // implemented yet!
            //
            endque( hp );

            /* klt - 10/30/96 - update active list */
            /* if we're in active feed mode        */
            if( active_feed )
            {
#if DEBUG_ACTIVE_FEED
                fprintf(stderr,"do_record() calling update_active_queue()\n");
#endif
                update_active_queue( hp );
            }

            return 0;
        }
        hp = hp->hr_clink.next;
    }

    return new_record( hashindex, &fd, timestamp, interface );
}

/*
 * new_record
 *  establish a new flow
 *  1> insert into hash table [maintained as stack]
 *  2> increment chain length
 *  3> insert at end of access-time queue
 */

int
new_record(
    u_int indx,                  /* bucket# in hash table */
    flowdata *fd,                /* record */
    CardClock timestamp,
    Bit8 interface
)
{
    struct hashrec *hr, *tmp;

    if ((hr = hash_free) == NULL)
    {
#if TALK_WHEN_ALLOC_FAILS
        fprintf(stderr, "unable to alloc new hashrec\n");
#endif
        hash_alloc_failures++;
        return 2;
    }
    hash_free = hr->hr_clink.next;

    if (pflag && !rflag) {
        double tmp = CardClock(timestamp - istamp).get_seconds();
        printf("N %.6f\t%x:%x\t0\t0\t0\t%d:%d:%d\t%.6f\n",
            tmp,
            fd->fd_src, fd->fd_dst, fd->fd_proto,
            fd->fd_sport, fd->fd_dport,
            timestamp.get_seconds()
        );
    }
    hr->hr_access_if = hr->hr_start_if = interface;
    hr->hr_access = hr->hr_start = timestamp;
    hr->hr_clink.next = hr->hr_clink.prev = NULL;
    hr->hr_alink.next = hr->hr_alink.prev = NULL;
    hr->hr_fd = *fd; /* copy from trace */
    hr->hr_bytes = Bit32( fd->fd_len );
    hr->hr_packets = 1;
    hr->hr_index = indx;

    insque( hr );

    total_flows++;
    u_int pending_flows = total_flows - deleted_flows;
    if( max_flows < pending_flows )
        max_flows = pending_flows;

#if DEBUG_CHAIN_LENGTH
    /*
     * increment chain length for this hash index
     */
    chainlen[indx]++;

#if DEBUG_HASHINDEX
    if( indx == hashindex_trace )
        fprintf( stderr, "hash index 0x%8.8lX chain length increased to %d\n", indx, chainlen[indx] );
#endif
#endif

    /* klt - 10/30/96 - add to active list */
    /* if we're in active feed mode        */
    if (active_feed)
    {
#if DEBUG_ACTIVE_FEED
        fprintf(stderr,"new_record() calling update_active_queue()\n");
#endif
        update_active_queue( hr );
    }

    /*
     * handle the case where no flows existed
     * on this hash chain until now
     */
    if ((tmp = flowtab[indx]) == NULL) {
        flowtab[indx] = hr;
        return 0;
    }

    /*
     * at least one flow already there (we collided earlier)
     * insert at the head [this maintains a stack]
     */

    hr->hr_clink.next = tmp;
    tmp->hr_clink.prev = hr;
    hr->hr_clink.prev = NULL;
    flowtab[indx] = hr;

#if DEBUG_CHAIN_LENGTH
    /*
     * if this is the longest chain we have ever seen,
     * update our global, and notify the user
     *
     * notifications should be asymptotically decreasing
     */
    if( longest_chain < chainlen[indx] )
    {
        longest_chain = chainlen[indx];
        fprintf( stderr, "hash index 0x%8.8lX is now longest chain at %d\n", indx, longest_chain );
        fprintf( stderr, "calced from IP=%8.8X:%8.8X proto=%d TCP=%4.4X:%4.4X\n", fd->fd_src, fd->fd_dst, fd->fd_proto, fd->fd_sport, fd->fd_dport );
    }
#endif

    return 0;
}

/*
 * delete_rec
 *  decrement chain length for that hash index
 *  and delete an entry from the hash table
 */

void delete_rec( struct hashrec *hr )
{
#if DEBUG_CHAIN_LENGTH
    u_int indx = hr->hr_index;

    chainlen[indx]--;
    if( chainlen[indx] < 0 )
    {
        fprintf( stderr, "warning: %d chain length for hash index 0x%8.8lX\n", chainlen[indx], indx );
    }

#if DEBUG_HASHINDEX
    if( indx == hashindex_trace )
        fprintf( stderr, "hash index 0x%8.8lX chain length decreased to %d\n", indx, chainlen[indx] );
#endif
#endif

    if (hr->hr_clink.prev != NULL)
        hr->hr_clink.prev->hr_clink.next = hr->hr_clink.next;
    else
        flowtab[hr->hr_index] = hr->hr_clink.next;

    if (hr->hr_clink.next != NULL)
        hr->hr_clink.next->hr_clink.prev = hr->hr_clink.prev;

    hr->hr_clink.next = hash_free;
    hash_free = hr;

    deleted_flows++;
}

/*
 * fill in seconds that may not have been printed due
 * to lack of trace data.  Makes sure we have at least some
 * trace record every housekeep_timeout*TVMULT secs
 */

void fillin_seconds( CardClock *ostamp, CardClock *stamp )
{
    CardClock tmp, addto, j;
    addto = TVMULT * (double) housekeep_timeout;
    tmp = *ostamp + addto;


#if DEBUG_FILLIN
    printf( "LOOPSTUFF: %g, %g\n", tmp.get_seconds(), stamp->get_seconds() );
#endif

    if (tmp < *stamp) {
        addto.put_seconds( 1 );
        for (j = tmp; j<=*stamp; j+=addto)
        {
            CardClock t;
            t = j - istamp;
            printf("F at+ %.6f sec, %d on entry, 0 new, 0 deleted, %d remaining, %d c\n",
                t.get_seconds(),
                total_flows - ndeleted_before_clean,
                total_flows - deleted_flows,
                flow_access_timeout.get_seconds()
            );
        }
    }
}



// called each time the housekeep_timeout has expired
//
void flows_expire( CardClock *now, Bit8 current_interface )
{
    struct hashrec *hp, *next;
    int ndeleted_this_time;
    Bit16 new_bytes_per_packet;
    u_int new_packets;
    Bit64 new_bytes;
    u_int new_flows;
    double new_houseclean_interval;
    double local_max_houseclean_interval = DBL_MIN;
    unsigned interface;
    unsigned bucket_count;

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        if( interface == current_interface )
            new_houseclean_interval = CardClock( *now - last_report_stamp[interface] ).get_seconds();
        else
            new_houseclean_interval = CardClock( end_block_stamp[interface] - last_report_stamp[interface] ).get_seconds();

        // longest interval is best estimate for non-per-card flows per second
        // (see assignment to new_flows after this loop, for more comments)
        //
        if( local_max_houseclean_interval < new_houseclean_interval )
            local_max_houseclean_interval = new_houseclean_interval;

    #if DELTA_TIME_PARANOIA
        if( new_houseclean_interval > 10.0 )
        {
            printf( "new_houseclean_interval %f is way too big!\n", new_houseclean_interval );
            printf( "CardClock *now high=%u low=%u card=0x%.4X\n", now->high, now->low, current_card );
            printf( "CardClock last high=%u low=%u card=0x%.4X\n", last_report_stamp[interface].high, last_report_stamp[interface].low, last_card );
            printf( "CardClock prev high=%u low=%u card=0x%.4X\n", prev_stamp.high, prev_stamp.low, prev_card );
            printf( "CardClock beg  high=%u low=%u card=0x%.4X\n", beg_block_stamp.high, beg_block_stamp.low, beg_block_card );
            printf( "CardClock end  high=%u low=%u card=0x%.4X\n", end_block_stamp[interface].high, end_block_stamp[interface].low, end_block_card );

    #if DELTA_TIME_FIXUP
            new_houseclean_interval = housekeep_timeout.get_seconds();
    #endif
        }
        else
    #endif
        {
    #if DELTA_TIME_STATS
            if( max_houseclean_interval < new_houseclean_interval )
                max_houseclean_interval = new_houseclean_interval;

            if( min_houseclean_interval > new_houseclean_interval )
                min_houseclean_interval = new_houseclean_interval;

            sum_houseclean_interval += new_houseclean_interval;
    #endif
        }

        // if any data has been received by this interface...
        //
        // otherwise, we want to avoid dividing by zero
        // and no point in adding zero to accumulators
        // and no point in comparing zero with maxes
        //
        if( new_houseclean_interval != 0.0 )
        {
            new_packets = (total_packets[ interface ] - cleaned_packets[ interface ]) / new_houseclean_interval;
            new_bytes = double(total_bytes[ interface ] - cleaned_bytes[ interface ]) / new_houseclean_interval;

            if( max_bytes[ interface ] < new_bytes )
            {
#if DEBUG_MAX_BPS
                fprintf( stderr, "%d max bytes per second increased from %.f to %.f\n", interface, (double)max_bytes[interface], (double)new_bytes );
                fprintf( stderr, "%d total=%.f cleaned=%.f interval=%g\n", interface, (double)total_bytes[interface], (double)cleaned_bytes[interface], new_houseclean_interval );
#endif

#if MAX_BPS_PARANOIA
                // if too many bytes per second for this (OC3c) interface...
                //
                if( (double)new_bytes > 17E6 )
                    fprintf( stderr, "DANGER!  %.f is too many bytes per second!\n", (double)new_bytes );
#endif
                max_bytes[ interface ] = new_bytes;
            }

            if( max_packets[ interface ] < new_packets )
                max_packets[ interface ] = new_packets;

            if( new_packets )
                new_bytes_per_packet = double( new_bytes ) / new_packets;
            else
                new_bytes_per_packet = Bit32( 0 );

            if( max_bytes_per_packet[ interface ] < new_bytes_per_packet )
                max_bytes_per_packet[ interface ] = new_bytes_per_packet;

            sum_bytes[ interface ] += new_bytes;
            sum_packets[ interface ] += new_packets;
            sum_bytes_per_packet[ interface ] += Bit32( new_bytes_per_packet );
        }
    }

    // this is better than assigning to new_flows when current_interface
    // is seen in loop above, because we want to use the best value for
    // interval when query is forced (current_interface >= MAX_ATM_NICS)
    // and flows were still seen since last expiration pass
    //
    // also, leaving the stack variable uninitialized tended to end up
    // with the value 4447288, which caused sum_new_flows to be way too
    // big, which caused variances in the ST line to be negative!
    //
    if( local_max_houseclean_interval > 0.0 )
        new_flows = (total_flows - cleaned_flows) / local_max_houseclean_interval;
    else
        new_flows = 0;

    if( max_new_flows < new_flows )
        max_new_flows = new_flows;

    sum_new_flows += new_flows;

    // small hack: delete_rec() will modify deleted_flows
    //
    ndeleted_before_clean = deleted_flows;

    // for all flows, least recently accessed first...
    //
    for( hp=Q.head; hp; hp=next )
    {
        // if this flow saw packets recently...
        //
        if( *now <= hp->hr_access + flow_access_timeout )

            // early out, because we have already seen any
            // less recently accessed flows, since they were
            // sorted to front of this list
            //
            break;

        // time to expire a flow
        //
        summarize_rec( "", &prefs, hp );

#if DEBUG_HOUSECLEAN
        dumphashrec( hp );
#endif

        // remember where the next flow is, since we are
        // about to alter this record's knowledge of it
        //
        next = hp->hr_qlink.next;

        // take this flow off the access time list
        //
        delque( hp );

        // take this flow off its collision-resolution list
        // and free up its memory
        //
        delete_rec( hp );
    }

    // for all hash chains...
    //
    for( bucket_count=0; bucket_count<NFLOWBUCKETS; bucket_count++ )
    {
        // for all hash records in this chain, least recently created first...
        //
        for( hp=flowtab[bucket_count]; hp; hp=next )
        {
            // if this flow was started recently enough...
            //
            if( *now <= hp->hr_start + flow_create_timeout )

                // early out, because we have already seen any
                // less recently created flows, since they were
                // sorted to front of this list
                //
                break;

            // time to expire a flow
            //
            summarize_rec( "", &prefs, hp );

#if DEBUG_HOUSECLEAN
            dumphashrec( hp );
#endif

            // take this flow off the access time list
            //
            delque( hp );

            // remember where the next flow is, since we are
            // about to alter this record's knowledge of it
            //
            next = hp->hr_clink.next;

            // take this flow off its collision-resolution list
            // and free up its memory
            //
            delete_rec( hp );
        }
    }

    ndeleted_this_time = deleted_flows - ndeleted_before_clean;
#if EXPIRE_OVERHEAD_STATS
    sum_deleted_flows += ndeleted_this_time;
    if( max_deleted_flows < ndeleted_this_time )
        max_deleted_flows = ndeleted_this_time;
#endif

    // pending flows stats
    //
    // note that we calculate max flows as flows are created,
    // not here, since it isn't a per-second measure and we want
    // it to reflect the true high-water mark, not a sampled one
    //

    sum_flows += total_flows - deleted_flows;

    if( fillin_wanted )
    {
        CardClock tmp = *now - istamp;
        fillin_seconds(&last_report_stamp[interface], now);

        printf("F at %.6f sec, %d on entry, %d new, %d deleted, %d remaining, %.6f tmo, %.6f\n",
            tmp.get_seconds(),
            total_flows - ndeleted_before_clean,
            total_flows - cleaned_flows,
            ndeleted_this_time,
            total_flows - deleted_flows,
            flow_access_timeout.get_seconds(),
            now->get_seconds()
        );
    }

    // we can't simply add the housekeeping timeout
    // to the previous report timestamp, because if we
    // are very far behind, it would still end up less
    // than current time, causing this function to be
    // called as soon as the next packet is seen
    //
    // that would be wasteful of CPU because we would iterate
    // through all flows with nearly no time elapsed and hence
    // nearly no flows to expire
    //
    // more importantly, if the link is busy on more than one
    // virtual circuit, the second packet could be right in
    // the middle of the first one, causing us to calculate
    // too small a value for new_houseclean_interval (remember
    // that the timestamp the card attached is for the first
    // cell only), which would inflate all our per-second byte
    // and packet counts, possibly way beyond what is possible
    // for the link!
    //
    next_report_stamp = *now + housekeep_timeout;

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        if( interface == current_interface )
            last_report_stamp[ interface ] = *now;
        else
            last_report_stamp[ interface ] = end_block_stamp[ interface ];

        cleaned_packets[ interface ] = total_packets[ interface ];
        cleaned_bytes[ interface ] = total_bytes[ interface ];
    }
    cleaned_flows = total_flows;
}



void usage( char *prog )
{
    fprintf(stderr, "Usage: %s [-pkh] [-o object]\n", prog );
    fprintf(stderr, "\t[-s create-timeout] [-t access-timeout]\n");
    fprintf(stderr,"\t[-a multicast-address-to-spew-active-flows-to]\n");
    fprintf(stderr,"\t[-u ip-address-to-tunnel-multicast-through]\n");
    fprintf(stderr,"\t[-l time-to-live-for-multicast-packets]\n");
    fprintf(stderr,"\t[-1 srcaddr] [-2 dstaddr] [-3 prot] [-4 sport] [-5 dport]\n");
    fprintf(stderr,"\t(p for printing, k for suppressing per second status \n");
    fprintf(stderr,"\th for help)\n");
}



int flows_init(
    int argc,
    char *argv[],
    PentiumClock *keystroke_interval,
    PentiumClock *query_interval,
    PentiumClock *active_interval
)
{
    u_short rsp, rdp, rpr;
    struct in_addr rsa, rda;
    unsigned count;
    int option_flag;
    int result;
    extern char *optarg;
    unsigned interface;

    if (argc < 2) {
        usage(argv[0]);
        return 1;
    }

    while ((option_flag = getopt(argc, argv, "pkh:l:u:a:o:s:t:1:2:3:4:5:")) != -1) {
        switch (option_flag) {

        /* klt - 10/30/96 - option "a:<mcast addr>" for active feed */
        case 'a':
            /* multicast feed address */
            maddr.s_addr = resolve( optarg );
            active_feed = TRUE;
            break;
        /* jza - 11/13/96 - option "u:<multicast tunnel addr>" for active feed */
        case 'u':
            /* multicast tunnel address */
            taddr.s_addr = resolve( optarg );
            break;
        /* jza - 11/13/96 - option "l:<multicast time to live>" for active feed */
        case 'l':
            /* multicast time to live */
            mttl = atoi( optarg );
            break;
        case 'o':
            objname = optarg;
            break;
        case 'p':
            pflag = TRUE;
            break;
        case 'k':
            fillin_wanted = FALSE;
            break;
        case 's':
            flow_create_timeout.put_seconds( atof( optarg ) );
            break;
        case 't':
            flow_access_timeout.put_seconds( atof( optarg ) );
            break;
        case '1':
            /* source address */
            sscanf(optarg, "%x", &rsa.s_addr);
            prefs.up_dst = rsa;
            rflag = TRUE;
            break;
        case '2':
            /* dest address */
            sscanf(optarg, "%x", &rda.s_addr);
            prefs.up_src = rda;
            rflag = TRUE;
            break;
        case '3':
            /* protocol number */
            rpr = atoi(optarg);
            break;
        case '4':
            /* source port */
            rsp = atoi(optarg);
            prefs.up_sport = rsp;
            break;
        case '5':
            /* dest port */
            rdp = atoi(optarg);
            prefs.up_dport = rdp;
            break;
        case '?':
        case 'h': /* help */
        default:
            usage(argv[0]);
            return 1;
        }
    }

    printf( "flows time out when created >%.1f seconds ago or accessed >%.1f seconds ago\n",
        flow_create_timeout.get_seconds(), flow_access_timeout.get_seconds() );
    printf( "flows pflag=%s\n", pflag?"TRUE":"FALSE" );
    printf( "flows fillin %swanted\n", fillin_wanted?"":"not " );
    printf( "flows query output buffer size is %u bytes\n", QUERY_BUFFER_SIZE );

    // sorry, we cannot init TCP/IP library
    // see main() in capture.cpp for reasons
    //

    // force all our TCP/IP sockets to a known state
    //
    memset( &query_sockets, 0, sizeof(query_sockets) );
    memset( &control_socket, 0, sizeof(control_socket) );

    // allocate output buffers for query sockets
    // so we get back to processing raw packets quickly,
    // even when the query connection is very slow
    //
    for( count=0; count<QUERY_SOCKET_COUNT; count++ )
    {
        query_output_buffers[ count ] = (Bit8 *) flat_alloc( QUERY_BUFFER_SIZE );
        if( query_output_buffers[ count ] == NULL )
        {
            printf( "error %d allocating %d bytes for query output buffer %d\n", errno, QUERY_BUFFER_SIZE, count );
            return 1;
        }
    }

    // listen for a TCP/IP connection from any host/port
    // on the remote control socket
    // without timing out, with no data handling callback
    // function and no busy-loop callback function
    //
    // no point in checking result; always succeeds
    //
    tcp_listen( &control_socket, CONTROL_TCP_PORT, 0, 0, NULL, 0 );

    // start using the first query socket
    //
    which_query_socket_to_use = 0;

    // no query is pending yet
    //
    which_query_socket_prev_used = QUERY_SOCKET_COUNT;
    query_drainage_delta = 0.0;

    /* klt - 10/30/96 - active feed */
    if (active_feed)
    {
        longword dport = ACTIVE_FEED_PORT;
        word iport;
        char maddr_string[ 17 ];
        char taddr_string[ 17 ];

        printf("Tunneling active flow updates to %s through %s with TTL %d\n",
            inet_ntoa(maddr_string, maddr.s_addr),
            inet_ntoa(taddr_string, taddr.s_addr),
            mttl);

        /* open a udp socket for sending it out */
        if (!udp_open(&msock,iport,maddr.s_addr,dport,NULL)) {
            printf("ERROR - unable to open udp socket\n");
            exit(1);
        }

#if DEBUG_ACTIVE_FEED
        printf("Opened UDP socket to send active feed to 0x%.8X\n", maddr.s_addr);
#endif
        sock_tunnel( &msock, taddr.s_addr );
        sock_ttl( &msock, mttl );

        // set an "alarm" time when we will
        // query the statistics
        //
        active_interval->put_seconds( ACTIVE_SPEW_MILLISECS / 1000 );
    }
    else
        printf( "flows not actively feeding anyone\n" );

    if (rflag) {
        printf("Tracing %x:%x\t%d:%d:%d\n", rsa.s_addr, rda.s_addr, rpr, rsp, rdp);
    }

    if ((objtype = findobj(objname)) < 0) {
        fprintf(stderr, "invalid option '%s' for -o\n", objname);
        fprintf(stderr, "valid types are: ");
        dumpobjtypes();
        return 1;
    }
    printf( "flows using object name '%s'\n", objname );




    fprintf( stderr, "buffer blocks are %d bytes each\n", sizeof(struct BufferBlock) );
    fprintf( stderr, "trace records are %d bytes each\n", sizeof(struct tracerec) );
    fprintf( stderr, "flow hash records are %d bytes each\n", sizeof(struct hashrec) );
    fprintf( stderr, "active flow stats are %d bytes each, so %d fit in each %d byte UDP payload\n",
        sizeof(struct flowstats), FLOWS_PER_UDP, UDP_PAYLOAD_SIZE );

    /* allocate buckets to point at each hash chain */
    flowtab = (struct hashrec **) flat_alloc( BUCKET_POOL_SIZE );
    if( flowtab == NULL )
    {
        printf( "error %d allocating %d bytes (%d pointers) for flow hash buckets\n", errno, BUCKET_POOL_SIZE, NFLOWBUCKETS );
        return 1;
    }

    /* nothing in any of the buckets yet */
    memset( flowtab, 0, BUCKET_POOL_SIZE );

#if DEBUG_CHAIN_LENGTH
    /* allocate chain length counters for each bucket */
    chainlen = (int *) flat_alloc( CHAIN_POOL_SIZE );
    if( chainlen == NULL )
    {
        printf( "error %d allocating %d bytes (%d integers) for chain length counters\n", errno, CHAIN_POOL_SIZE, NFLOWBUCKETS );
        return 1;
    }

    /* initialize chain length of all hashtable buckets to zero
     */
    for( count=0; count<NFLOWBUCKETS; count++ )
        chainlen[ count ] = 0;
#endif

    /* allocate a pool of hash records */
    hash_pool = (struct hashrec *) flat_alloc( HASH_POOL_SIZE );
    if( hash_pool == NULL )
    {
        printf( "error %d allocating %d bytes (%d records) for flow hash record pool\n", errno, HASH_POOL_SIZE, HASH_RECORDS );
        return 1;
    }

    /* capture module should ask me about query every chance it gets */
    query_interval->put_seconds( 0 );

    /* capture module only needs to check for keystrokes once a second */
    keystroke_interval->put_seconds( 1.0 );

    /* put all the hash records on the singly-linked free list */
    hash_free = hash_pool;
    for( count=0; count<HASH_RECORDS-1; count++ )
    {
        hash_pool[ count ].hr_clink.next = hash_pool + count + 1;
    }
    hash_pool[count].hr_clink.next = NULL;

    // record program start time (longer-term than
    // this single collection interval)
    //
    // note: this may be irrelevant because main data
    // capture program is probably already doing this
    //
#if __unix__
    /* get the current CPU usage so we can subtract
     * it out of the final number, so we can ignore
     * disk I/O
     */
    getrusage( RUSAGE_SELF, &before );
#else
    before.set();
#endif

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        cleaned_bytes[ interface ] = Bit32( 0 );
        cleaned_packets[ interface ] = 0;
        total_bytes[ interface ] = Bit32( 0 );
        total_packets[ interface ] = 0;
        end_block_stamp[ interface ] = 0;
        last_report_stamp[ interface ] = 0;
        last_already_set[ interface ] = FALSE;
    }
    report_already_set = FALSE;
    cleaned_flows = 0;
    deleted_flows = 0;
    total_flows = 0;
    query_count = -1;

    return flows_clear();
}



// clears all statistics after being queried remotely
// or during initialization of module
//
// note that we do not touch the report timestamp
// so that querying will not delay timing out a flow
//
int flows_clear( void )
{
    int result;
    unsigned interface;

    /* initialize 2nd-level stats
     */
    result = init_stats();
    if( result )
        return result;

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
        sum_bytes[ interface ] = Bit32( 0 );
        max_bytes[ interface ] = Bit32( 0 );
        instant_max_bytes[ interface ] = 0;
        sum_packets[ interface ] = 0;
        max_packets[ interface ] = 0;
        sum_bytes_per_packet[ interface ] = Bit32( 0 );
        max_bytes_per_packet[ interface ] = 0;
        non_ip_packets[ interface ] = 0;
        ip_optioned_packets[ interface ] = 0;
        non_first_ip_fragments[ interface ] = 0;
        first_ip_fragments[ interface ] = 0;
    }
    max_flows = 0;
    sum_flows = 0;
    max_new_flows = 0;
    sum_new_flows = 0;
#if EXPIRE_OVERHEAD_STATS
    max_deleted_flows = 0;
    sum_deleted_flows = 0;
#endif

#if DELTA_TIME_STATS
    sum_houseclean_interval = 0;
    max_houseclean_interval = DBL_MIN;
    min_houseclean_interval = DBL_MAX;
#endif

    hash_alloc_failures = 0;
    istamp_already_set = FALSE;
    query_count++;

    memset( ip_length_usage, 0, sizeof(ip_length_usage) );
    memset( ip_precedence_usage, 0, sizeof(ip_precedence_usage) );

    return 0;
}



// called in foreground by main data capture program
//
int flows_per_block( BufferBlock *block )
{
    int final_result = 0;
    int result;
    unsigned count;
    Bit8 interface = block->cardindex;
    tracerec *record;
    union { FpuStatus float_status; Bit16 float_status_alias; };
#if DEBUG_COPY_PACKET
    tracerec debug_record;
#endif

#if BUFFER_USED_PARANOIA
    // safety check: used count cannot be greater
    // than COMPL_RING_ENTRIES
    //
    if( block->used > COMPL_RING_ENTRIES )
    {
        printf( "internal error: block used %d > %d max!\n", block->used, COMPL_RING_ENTRIES );
        return 1;
    }
#endif

    /* for all packets headers in the block
     */
    for( count=block->queried; count<block->used; count++ )
    {
#if BUFFER_PTR_PARANOIA
        // safety check: buffer cannot be before
        // the start of the flat memory pool
        //
        if( block->buffers < flat_pool )
        {
            printf( "danger!  corrupted bufferblock!  stop here in debugger\n" );
            return 2;
        }
#endif

#if FLOAT_EXCEPT_PARANOIA
        // if there was any exception but precision...
        //
        float_status_alias = get_fpu_status();
        if( float_status_alias & 0xDF )
            printf( "before record %d floating point status is 0x%.4X\n", count, float_status_alias );
        clear_fpu_status();
#endif

        record = (tracerec *) (block->buffers + count*BUFFER_SIZE);

#if DEBUG_COPY_PACKET
        // copy the record because the debugger, for some
        // idiotic reason does not have the same addressibility
        // that our program has
        //
        debug_record = *record;
#endif

        total_packets[ interface ]++;

        /* shortcut variable
         */
        tstamp = record->when;
        tstamp.strip_bitfields();

#if DELTA_TIME_PARANOIA
        if( count == 0 )
        {
            beg_block_card = block->cardnum;
            beg_block_stamp = tstamp;
        }
#endif

        /* if we haven't seen any data since last query...
         */
        if( !istamp_already_set )
        {
            istamp_already_set = TRUE;

            // remember when the first packet was seen
            //
            istamp = tstamp;
        }

        // if we never set the report timestamp...
        //
        if( !report_already_set )
        {
            report_already_set = TRUE;

            // do housekeeping one timeout from now
            //
            next_report_stamp = tstamp + housekeep_timeout;
        }

        // if we haven't seen any data on this interface yet...
        //
        if( !last_already_set[ interface ] )
        {
            last_already_set[ interface ] = TRUE;

            last_report_stamp[ interface ] = tstamp;
            end_block_stamp[ interface ] = tstamp;
        }

        /* report results if the interval is correct
         * (each 'housekeep_timeout' time quanta)
         *
         * no point in doing this unless the timestamp
         * could have changed, outside the
         */
        if( tstamp >= next_report_stamp )
        {
#if DELTA_TIME_PARANOIA
            current_card = block->cardnum;
#endif

            flows_expire( &tstamp, interface );

#if DELTA_TIME_PARANOIA
            last_card = block->cardnum;
#endif
        }

#if DELTA_TIME_PARANOIA
        prev_stamp = tstamp;
        prev_card = block->cardnum;
#endif

#if !ASSUME_ETHERTYPE_IS_IP
        /* if MAC protocol is IP...
         *
         * assumes that if LLC or SNAP OUI are wrong,
         * it is likely that ethertype is also wrong
         */
        if( record->tr_etype != htons(ETHERTYPE_IP) )
        {
            non_ip_packets[ interface ]++;
            continue;
        }
#endif

        record->tr_len=ntohs(record->tr_len);
        total_bytes[ interface ] += Bit32( record->tr_len );
        if( instant_max_bytes[ interface ] < record->tr_len )
            instant_max_bytes[ interface ] = record->tr_len;
        ip_length_usage[ record->tr_len >> IP_LEN_SHIFT ]++;

        ip_precedence_usage[ (record->tr_tos >> IP_PREC_SHIFT) & IP_PREC_MASK ]++;

        record->tr_offset = ntohs( record->tr_offset );

        /* if object type (statistic to track) is "portquad"
         * then we must toss packets with options and fragments
         * that are not the first
         */
#if !ASSUME_OBJECT_IS_PQ
        if( strcmp(objname,"pq") == 0 )
#endif
        {
            if( record->tr_ip.ip_hl != (sizeof(ip) >> 2) )
            {
                ip_optioned_packets[ interface ]++;
                continue;
            }

            if( (record->tr_offset & IP_OFF_MASK) != 0 )
            {
                non_first_ip_fragments[ interface ]++;
                continue;
            }

            if( record->tr_offset & IP_MF )
                first_ip_fragments[ interface ]++;
        }

#if BYTESWAP_FIELDS_EARLY
        record->tr_src=ntohl(record->tr_src);
        record->tr_dst=ntohl(record->tr_dst);
        record->tr_sport=ntohs(record->tr_sport);
        record->tr_dport=ntohs(record->tr_dport);
#endif

        /* insert new record */
        result = do_record( record, tstamp, interface );
        if( result )
            if( final_result == 0 )
                final_result = result;

    }   // end for all records in block

    // remember how much of this block we queried, in
    // case it is a partial that we will see again later
    //
    block->queried = count;

    end_block_stamp[ interface ] = tstamp;
#if DELTA_TIME_PARANOIA
    end_block_card = interface;
#endif

    return final_result;
}



// called to determine if program has been queried remotely
//
int flows_is_query_wanted( void )
{
    int result;

    // give the TCP/IP library some time
    //
    tcp_Socket *current_query_socket = &query_sockets[ which_query_socket_to_use ];
    Bit8 *current_query_buffer = query_output_buffers[ which_query_socket_to_use ];
    result = tcp_tick( current_query_socket );

    // if a query is draining and that previous socket has been closed...
    //
    if( which_query_socket_prev_used < QUERY_SOCKET_COUNT
    &&  !_chk_socket( query_sockets + which_query_socket_prev_used ) )
    {
        // remember that no query is draining anymore
        //
        which_query_socket_prev_used = QUERY_SOCKET_COUNT;

        // calculate how long it took to push the stats
        // through the damned sluggish network
        //
        PentiumClock now;
        now.set();
        query_drainage_delta = PentiumClock( now - query_start_time ).get_seconds();

        // put it on the console for debugging
        //
        printf( "previous query socket took %.3g seconds to drain through network\n", query_drainage_delta );
    }

    // if current socket has been closed...
    //
    if( !result )
    {
        // listen for a TCP/IP connection from any host/port,
        // without timing out, with no data handling callback
        // function and no busy-loop callback function
        //
        // no point in checking result; always succeeds
        //
        tcp_listen( current_query_socket, QUERY_TCP_PORT, 0, 0, NULL, 0 );

        // attach a much-larger output buffer to the socket
        //
        if( 0 == sock_set_tx_buf( current_query_socket, current_query_buffer, QUERY_BUFFER_SIZE ) )
            printf( "error: could not set TX buf for socket #%d\n", which_query_socket_to_use );

        return FALSE;
    }

    // otherwise if someone has connected to us
    //
    if( sock_established( current_query_socket ) )
    {
        // treat this as a query
        //
        // I know this is crude, but I don't want to
        // implement a whole request/response protocol yet
        //
        return TRUE;
    }

    return FALSE;
}



// called when program is queried remotely
//
int flows_query( const char *extra_msg )
{
    double trace_duration;
    int result;
    unsigned interface;
    unsigned count;
    tcp_Socket *sock = &query_sockets[ which_query_socket_to_use ];
#if REPORT_OVERHEAD_STATS
    PentiumClock start, now;
    double elapsed, elapsed2;
#endif

    // remember when the query started, so we
    // can calculate network throughput later
    //
    query_start_time.set();

#if QUERY_CALLS_EXPIRE
#if DELTA_TIME_PARANOIA
    current_card = -1;
#endif

    // do one more housecleaning to trigger a timeout
    // on any "pregnant" flows, in case we haven't seen
    // a packet in more than housekeep_timeout ticks
    //
    // this fake timestamp will not cause a negative
    // delta time to be calculated during next flows_expire(),
    // even if it is later than the next packet we will see,
    // because flows_clear() will cause last_report_stamp
    // to be recalculated when we see the next packet
    //
    CardClock now2 = query_start_time;
    flows_expire( &now2, -1 );

#if DELTA_TIME_PARANOIA
    last_card = -1;
    prev_card = -1;
    prev_stamp = now2;
#endif
#endif      // QUERY_CALLS_EXPIRE

    fprintf( stderr, "\n" );
    for( interface=0; interface<MAX_ATM_NICS; interface++ )
    {
#if DEBUG_QUERY_TOTALS
        fprintf( stderr, "interface %u %u/%u/%u/%u/%u all/non-IP/IP-opt/frag-non-1/frag-1 packets\n",
            interface,
            sum_packets[interface],
            non_ip_packets[interface],
            ip_optioned_packets[interface],
            non_first_ip_fragments[interface],
            first_ip_fragments[interface]
        );

        fprintf( stderr, "interface %u total bytes: %.f\n", interface, (double)total_bytes[interface] );
#endif

#if FAIL_ON_PACKET_COUNT
        if( sum_packets[interface] < EXPECTED_PACKETS )
        {
            printf( "fatal error: expected at least %d packets on interface %u!\n", EXPECTED_PACKETS, interface );
#if WAIT_FOR_KEYSTROKE
            printf( "press a key and I will exit\n" );
            getch();
#endif
            return 1;
        }
#endif
    }
    fprintf( stderr, "%d flows started, %d deleted before cleanup\n", total_flows, deleted_flows );

    trace_duration = CardClock( tstamp - istamp ).get_seconds();





    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u IP bytes: %u\n", interface, sum_bytes[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u all packets: %u\n", interface, sum_packets[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u non IP packets: %u\n", interface, non_ip_packets[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u IP optioned packets: %u\n", interface, ip_optioned_packets[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u first IP fragments: %u\n", interface, first_ip_fragments[interface] );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u non-first IP fragments: %u\n", interface, non_first_ip_fragments[interface] );

    sock_printf( sock, "flow allocation failures: %u\n", hash_alloc_failures );
    sock_printf( sock, "port flow allocation failures: %u\n", porthash_alloc_failures );
    sock_printf( sock, "net flow allocation failures: %u\n", nethash_alloc_failures );
    sock_printf( sock, "host flow allocation failures: %u\n", hosthash_alloc_failures );

    sock_printf( sock, "following statistics were calculated at flow expiration time:"
        " (first packet >1 second after prev flow expiration)\n" );

    sock_printf( sock, "pending flows: %d max, %.0f avg\n",
        max_flows, sum_flows/nonzero(trace_duration) );

    sock_printf( sock, "new flows per second: %d max, %.0f avg\n",
        max_new_flows, sum_new_flows/nonzero(trace_duration) );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u packets per second: %d max, %.0f avg\n",
            interface, max_packets[interface], sum_packets[interface]/nonzero(trace_duration) );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u bits per second: %.0f max, %.0f avg\n",
            interface, 8*double(max_bytes[interface]), 8*double(sum_bytes[interface])/nonzero(trace_duration) );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u avg bytes per packet: %d max, %.0f avg\n",
            interface, max_bytes_per_packet[interface], double(sum_bytes_per_packet[interface])/nonzero(trace_duration) );

    // print out the counts for all IP types of service we have seen
    //
    sock_printf( sock, "IP_PREC" );
    for( count=0; count<IP_PREC_COUNT; count++ )
        if( ip_precedence_usage[count] > 0 )
            sock_printf( sock, " %d:%d", count, ip_precedence_usage[count] );
    sock_printf( sock, "\n" );

    for( interface=0; interface<MAX_ATM_NICS; interface++ )
        sock_printf( sock, "interface %u largest packet: %u bytes\n", interface, instant_max_bytes[interface] );

    // print out the counts for all IP lengths we have seen
    //
    sock_printf( sock, "IP_LEN range=%d", 1<<IP_LEN_SHIFT );
    for( count=0; count<IP_LEN_COUNT; count++ )
        if( ip_length_usage[count] > 0 )
            sock_printf( sock, " %d:%d", count<<IP_LEN_SHIFT, ip_length_usage[count] );
    sock_printf( sock, "\n" );

#if REPORT_OVERHEAD_STATS
    start.set();
#endif

    // output the stats that got calculated for expired flows
    //
    result = report_stats( sock, istamp.get_seconds(), trace_duration );

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( start - query_start_time ).get_seconds();
    elapsed2 = PentiumClock( now - query_start_time ).get_seconds();
    fprintf( stderr, "cumulative %.4f S before, %.4f after report_stats()\n", elapsed, elapsed2 );
    sock_printf( sock, "cumulative %.4f S before, %.4f after report_stats()\n", elapsed, elapsed2 );
#endif

    if( result )
        return result;

#if DELTA_TIME_STATS
    sock_printf( sock, "seconds since last flow housecleaning: %f min, %f max, %f avg\n",
        min_houseclean_interval, max_houseclean_interval,
        sum_houseclean_interval/nonzero(trace_duration) );
#endif

#if EXPIRE_OVERHEAD_STATS
    sock_printf( sock, "deleted flows: %d max, %.1f avg\n",
        max_deleted_flows, sum_deleted_flows/nonzero(trace_duration) );
#endif

    // say how long it took to push the stats through
    // the network last time
    //
    sock_printf( sock, "%.3g seconds for previous query socket to drain"
        " through network\n", query_drainage_delta );

    // append the caller's message to the socket
    //
    if( extra_msg )
        sock_puts( sock, (byte *)extra_msg );

#if REPORT_OVERHEAD_STATS
    start.set();
#endif

    result = flows_clear();
    if( result )
        return result;

#if REPORT_OVERHEAD_STATS
    now.set();
    elapsed = PentiumClock( now - start ).get_seconds();
    fprintf( stderr, "clear:%.4f\n", elapsed );
    sock_printf( sock, "clear:%.4f\n", elapsed );
#endif

    // close the TCP/IP connection from this end
    //
    // when caller closes or we time out, we will
    // know to listen again
    //
    sock_close( sock );

    // move on to the next socket, so somebody can
    // connect to us with a different TCP source port
    // before TIME_WAIT elapses on the old TCP port#
    //
    which_query_socket_prev_used = which_query_socket_to_use;
    which_query_socket_to_use++;
    if( which_query_socket_to_use >= QUERY_SOCKET_COUNT )
        which_query_socket_to_use = 0;

    return 0;
}



/*
 * summarize_rec
 *  print out fascinating detail about hashrec
 */
void summarize_rec(
    char *cp,
    struct user_prefs *up,
    struct hashrec *hr
)
{
    CardClock tmp, tmp2;
    float   duration;

    /* if rflag set, and no match, early out */
    if (rflag && !flowcmp(&up->up_fd, &hr->hr_fd))
        return;

    tmp = hr->hr_access - hr->hr_start;
    duration = tmp.get_seconds();
#if 0
    // joel debug
    if( (int)tmp.high < 0 )
    {
        printf( "error: duration %.f > 1000\n", duration );
        dumphashrec( hr );
    }
#endif

    if (pflag) {
        tmp2 = tstamp - istamp;
        printf(
           "A%s: %.6f\t%x:%x\t%d\t%d\t%.3f\t%d:%d:%d\t%.6f\n",
            cp,
            tmp2.get_seconds(),
            hr->hr_fd.fd_src.s_addr,
            hr->hr_fd.fd_dst.s_addr,
            hr->hr_packets,
            hr->hr_bytes,
            duration,
            hr->hr_fd.fd_proto,
            hr->hr_fd.fd_sport,
            hr->hr_fd.fd_dport,
            tstamp.get_seconds()
        );
    }

#if !STUB_2ND_LEVEL_STATS
    update_stats( hr, duration );
#endif
}



/*
 *  this is called when EOF is reached
 *  to walk through still-pending flows
 *  and force them to be accumulated statistically
 */
void summarize( void )
{
    extern struct queue Q;
    struct hashrec *hp = Q.head;
    struct hashrec *next;

    while (hp != NULL) {
        summarize_rec("-EXIT", &prefs, hp);

        next = hp->hr_qlink.next;
        delque(hp);
        delete_rec(hp);

        hp = next;
    }
}



// called when program is about to exit
//
int flows_cleanup( void )
{
    summarize();

    // since nobody has called us when we are cleaning up,
    // there is no point in doing one last query, so this
    // code is commented out:
    //
    //     int result = flows_query( NULL );
    //
    int result = 0;

#if __unix__
    getrusage( RUSAGE_SELF, &after );
    after.ru_utime -= before.ru_utime;
    after.ru_stime -= before.ru_stime;
    fprintf( stderr, "after first disk I/O, we used %.3f user, %.3f system CPU seconds\n",
        after.ru_utime.tv_sec + after.ru_utime.tv_usec/TIME_DIVISOR,
        after.ru_stime.tv_sec + after.ru_stime.tv_usec/TIME_DIVISOR
    );
#else
    after.set();
    double elapsed = PentiumClock(after - before).get_seconds();
    fprintf( stderr, "not counting disk I/O, %g seconds elapsed\n", elapsed );
#endif

    return result;
}



/*
 * klt - 10/31/96
 * sends data on all hashrecs in "A" queue
 * if active feed mode is on
 */
void flows_send_active( void )
{
    /*printf("DEBUG --> active flows record...\n");*/
    /*dumphashrec(hp);*/
    struct hashrec *hpp, *next;
    extern struct queue A;
    static struct flowstats packet[ FLOWS_PER_UDP ];
    unsigned packet_fullness = 0;

    /* rifle thru A */
#if DEBUG_ACTIVE_FEED
    printf("Process thru A queue...\n");
#endif
    for( hpp=A.head; hpp; hpp=next )
    {
        next = hpp->hr_alink.next;

#if DEBUG_ACTIVE_FEED_DATA
#if BYTESWAP_FIELDS_EARLY
        printf("%X --> %X:", hpp->hr_fd.fd_src.s_addr, hpp->hr_fd.fd_dst.s_addr);
#else
        printf("%X --> %X:", ntohl(hpp->hr_fd.fd_src.s_addr), ntohl(hpp->hr_fd.fd_dst.s_addr));
#endif
        printf("%d pkts, %d bytes,",hpp->hr_packets,hpp->hr_bytes);
#if BYTESWAP_FIELDS_EARLY
        printf("%d sport, %d dport,",hpp->hr_fd.fd_sport,hpp->hr_fd.fd_dport);
#else
        printf("%d sport, %d dport,",ntohs(hpp->hr_fd.fd_sport),ntohs(hpp->hr_fd.fd_dport));
#endif
        printf("proto: %d\n",hpp->hr_fd.fd_proto);
#endif

        /* pack the flowstats into the current packet */
        packet[ packet_fullness++ ] = hpp->hr_fs;

        /* output packet if it is full */
        if( packet_fullness >= FLOWS_PER_UDP )
        {
            sock_write(&msock, (byte *)&packet, packet_fullness*sizeof(struct flowstats));
#if DEBUG_ACTIVE_FEED
            printf("wrote %d flowstats to UDP socket\n", packet_fullness);
#endif
            packet_fullness = 0;
        }

        /* remove it from A queue */
        del_active_queue(hpp);
    }

    /* flush packet if it is not empty */
    if( packet_fullness > 0 )
    {
        sock_write(&msock, (byte *)&packet, packet_fullness*sizeof(struct flowstats));
#if DEBUG_ACTIVE_FEED
        printf("wrote %d trailing flowstats to UDP socket\n", packet_fullness);
#endif
    }
}



// turn off compiler warning about unused parameters
// since I know they won't be used and I must have
// them because calling code cannot be changed
//
#pragma warn -par

// this code doesn't have anything special to do
// when keys are pressed
//
void flows_keystroke( int keystroke )
{
}
// turn compiler warning about unused parameters
// back to whatever it was before I messed with it
//
#pragma warn .par

#endif  // DO_FOREGROUND_STATS



/*
 * dumphashrec: diagnostic
 */
void dumphashrec( struct hashrec *hp )
{
    fprintf(stderr, "  #packets: %d\n", hp->hr_packets);
    fprintf(stderr, "  #bytes: %d\n", hp->hr_bytes);
    fprintf(stderr, "  start: %d %g\n", hp->hr_start_if, hp->hr_start.get_seconds() );
    fprintf(stderr, "  access: %d %g\n", hp->hr_access_if, hp->hr_access.get_seconds() );
    fprintf(stderr, "  srcIP: %X, dstIP: %X\n", hp->hr_fd.fd_src.s_addr, hp->hr_fd.fd_dst.s_addr);
    fprintf(stderr, "  proto: %d, sport: %d, dport: %d\n", hp->hr_fd.fd_proto, hp->hr_fd.fd_sport, hp->hr_fd.fd_dport);
}



